#pragma once

#pragma warning(push)
#pragma warning(disable : 4267) // conversion from a to b, possible loss of data
#pragma warning(disable : 4505) // unreferenced local function has been removed
#pragma warning(disable : 4702) // unreachable code

#include "register_function.h"
#include "register_class.h"
#include "register_property.h"
#include "public_util.h"
#include "dukvalue.h"

#pragma warning(pop)
